/* hash_update.h - functions to update a crc file */
#ifndef HASH_UPDATE_H
#define HASH_UPDATE_H

#ifdef __cplusplus
extern "C" {
#endif

int update_hash_file(file_t* file);

#ifdef __cplusplus
} /* extern "C" */
#endif /* __cplusplus */

#endif /* HASH_UPDATE_H */
